import React, { Component, MouseEvent, ChangeEvent } from 'react';
import { TodoItem } from './todoitem';
import { parseListResponse, parseUpdateResponse } from './response';
import './TodoApp.css';


type TodoState = {
  items: Array<TodoItem> | undefined;  // to-do items or undefined if still loading
  newName: string;                     // mirrors text in the field to add a new name
};


export class TodoApp extends Component<{}, TodoState> {

  constructor(props: {}) {
    super(props);

    this.state = {items: undefined, newName: ""};
  }

  componentDidMount = () => {
    fetch("/api/list")
        .then(this.doListResp)
        .catch(() => this.doListError("failed to connect to server"));
  }

  render = (): JSX.Element => {
    if (this.state.items === undefined) {
      return <div>
          <h2>To-Do List</h2>
          <p><i>Loading</i> the to-do list...</p>
        </div>;

    } else if (this.state.items.length === 0) {
      return (
        <div>
          <h2>To-Do List</h2>
          <p><i>No items</i> in the list.</p>
          <p>New item:
            <input type="text" value={this.state.newName}
                onChange={this.doNewNameChange} />
            <button type="button"
                onClick={this.doAddClick}>Add</button>
          </p>
        </div>);

    } else {
      return (
        <div>
          <h2>To-Do List</h2>
          {this.renderItems(this.state.items)}
          <p>Check the item to mark it completed. Check it again to remove it.</p>
          <p>New item:
            <input type="text" value={this.state.newName}
                onChange={this.doNewNameChange} />
            <button type="button"
                onClick={this.doAddClick}>Add</button>
          </p>
        </div>);
    }
  }

  renderItems = (items: Array<TodoItem>): Array<JSX.Element> => {
    const itemElems: Array<JSX.Element> = [];
    for (let i = 0; i < items.length; i++) {
      const item = items[i];
      if (item.completed) {
        itemElems.push(
          <div key={i}>
            <input type="checkbox" id={"check" + i} checked={true}
                onChange={evt => this.doItemRemove(evt, i)}/>
            <label htmlFor={"check" + i}><s>{item.name}</s></label>
          </div>);
      } else {
        itemElems.push(
          <div key={i}>
            <input type="checkbox" id={"check" + i} checked={false}
                onChange={evt => this.doItemCompleted(evt, i)} />
            <label htmlFor={"check" + i}>{item.name}</label>
          </div>);
      }
    }
    return itemElems;
  };

  doNewNameChange = (evt: ChangeEvent<HTMLInputElement>) => {
    this.setState({newName: evt.target.value});
  };

  doListResp = (res: Response): void => {
    if (res.status === 200) {
      res.json().then(this.doListJson)
         .catch(() => this.doListError("200 response is not valid JSON"));
    } else {
      res.text().then(this.doListError)
         .catch(() => this.doListError("error response is not text"));
    }
  };

  doListJson = (val: unknown): void => {
    this.setState({items: parseListResponse(val)});
  };

  doListError = (msg: string): void => {
    console.error(`Error fetching /api/list: ${msg}`);
  };

  doAddClick = (_evt: MouseEvent<HTMLButtonElement>) => {  // _evt since unused
    // Ignore the request if the user hasn't entered a name.
    const name = this.state.newName.trim();
    if (name.length == 0)
      return;

    const args = {name: name};
    fetch("/api/add", {
        method: "POST", body: JSON.stringify(args),
        headers: {"Content-Type": "application/json"}})
      .then(this.doAddResp)
      .catch(() => this.doAddError("failed to connect to server"));
  };

  doAddResp = (res: Response): void => {
    if (res.status === 200) {
      res.json().then(this.doAddJson)
         .catch(() => this.doAddError("200 response is not valid JSON"));
    } else {
      res.text().then(this.doAddError)
         .catch(() => this.doAddError("error response is not text"));
    }
  };

  doAddJson = (val: unknown): void => {
    const res = parseUpdateResponse(val);

    if (this.state.items === undefined)
      throw new Error('items list is undefined. impossible!');

    const items = this.state.items.concat([{name: res.name, completed: false}]);
    this.setState({items: items, newName: ""});
  };

  doAddError = (msg: string): void => {
    console.error(`Error fetching /api/add: ${msg}`);
  };

  doItemCompleted = (_evt: ChangeEvent<HTMLInputElement>, index: number) => {
    if (this.state.items === undefined)
      throw new Error('items list is undefined. impossible!');

    const item = this.state.items[index];
    const args = {name: item.name};
    fetch("/api/complete", {
        method: "POST", body: JSON.stringify(args),
        headers: {"Content-Type": "application/json"}})
      .then((res) => this.doCompleteResp(res, index))
      .catch(() => this.doCompleteError("failed to connect to server"));

  };

  doCompleteResp = (res: Response, index: number): void => {
    if (res.status === 200) {
      res.json().then((val) => this.doCompleteJson(val, index))
         .catch(() => this.doCompleteError("200 response is not valid JSON"));
    } else {
      res.text().then(this.doCompleteError)
         .catch(() => this.doCompleteError("error response is not text"));
    }
  };

  doCompleteJson = (val: unknown, index: number): void => {
    parseUpdateResponse(val);  // make sure it was a valid response

    if (this.state.items === undefined)
      throw new Error('items list is undefined. impossible!');

    const item = this.state.items[index];
    const items = this.state.items.slice(0, index)
        .concat([{name: item.name, completed: true}])
        .concat(this.state.items.slice(index + 1));
    this.setState({items: items});
  };

  doCompleteError = (msg: string): void => {
    console.error(`Error fetching /api/complete: ${msg}`);
  };

  doItemRemove = (_evt: ChangeEvent<HTMLInputElement>, index: number) => {
    if (this.state.items === undefined)
      throw new Error('there are no items!');

    const item = this.state.items[index];
    const args = {name: item.name};
    fetch("/api/remove", {
        method: "POST", body: JSON.stringify(args),
        headers: {"Content-Type": "application/json"}})
      .then((res) => this.doRemoveResp(res, index))
      .catch(() => this.doRemoveError("failed to connect to server"));
  };

  doRemoveResp = (res: Response, index: number): void => {
    if (res.status === 200) {
      res.json().then((val) => this.doRemoveJson(val, index))
         .catch(() => this.doRemoveError("200 response is not valid JSON"));
    } else {
      res.text().then(this.doRemoveError)
         .catch(() => this.doRemoveError("error response is not text"));
    }
  };

  doRemoveJson = (val: unknown, index: number): void => {
    parseUpdateResponse(val);  // make sure it was a valid response

    if (this.state.items === undefined)
      throw new Error('items list is undefined. impossible!');

    const items = this.state.items.slice(0, index)
        .concat(this.state.items.slice(index + 1));
    this.setState({items: items});
  };

  doRemoveError = (msg: string): void => {
    console.error(`Error fetching /api/remove: ${msg}`);
  };
}